from __future__ import annotations

from pathlib import Path
from typing import Optional

import typer

from lander.plugins import parsers, themes
from lander.settings import BuildSettings

__all__ = ["app"]

app = typer.Typer()


@app.command()
def build(
    output: Optional[Path] = typer.Option(
        None, help="Directory for the built site"
    ),
    source: Optional[Path] = typer.Option(
        None, help="Source file for metadata (usually the root tex file)."
    ),
    pdf: Optional[Path] = typer.Option(
        None, help="Path to the PDF file to display on the landing page."
    ),
    parser: Optional[str] = typer.Option(None, help="Metadata parsing plugin"),
    theme: Optional[str] = typer.Option(None, help="Theme plugin."),
    canonical_url: Optional[str] = typer.Option(
        None, "--url", help="Canonical URL where the landing page is hosted."
    ),
) -> None:
    settings = BuildSettings.load(
        output_dir=output,
        source_path=source,
        pdf=pdf,
        parser=parser,
        theme=theme,
        canonical_url=canonical_url,
    )

    # Load plugins
    Parser = parsers[settings.parser]
    Theme = themes[settings.theme]

    # Run the document parser
    parser_plugin = Parser(settings=settings)
    theme_plugin = Theme(metadata=parser_plugin.metadata, settings=settings)
    theme_plugin.build_site()

    print(f"Generated landing page in: {settings.output_dir}")


@app.command("themes")
def list_themes() -> None:
    print("Available themes:\n")
    print(themes.names)


@app.command("parsers")
def list_parsers() -> None:
    print("Available parsers:\n")
    print(parsers.names)



from __future__ import annotations

import mimetypes
from pathlib import Path
from typing import Any, Dict, List, Optional

import yaml
from pydantic import BaseModel, Field, FilePath, HttpUrl, validator

from lander.plugins import parsers, themes

__all__ = ["BuildSettings"]


class DownloadableFile(BaseModel):
    """Model for a file that's downloadable through the landing page, either
    the main PDF or an attachment.
    """

    file_path: FilePath
    """Path to the file on the filesystem (not the generated site)."""

    name: str
    """Name of the file."""

    mimetype: str
    """The mimetype of the file."""

    extension: str
    """The file's extension."""

    size: int
    """The file's size in bytes."""

    @classmethod
    def load(cls, path: Path) -> DownloadableFile:
        file_type, file_encoding = mimetypes.guess_type(
            str(path), strict=False
        )

        return cls(
            file_path=path,
            name=path.name,
            mimetype=file_type,
            extension=path.suffix,
            size=path.stat().st_size,
        )

    @property
    def human_size(self) -> str:
        """Humanized size label."""
        if self.size < 1000:
            return f"{self.size} B"
        elif self.size < 1000000:
            return f"{self.size / 1000:.0f} kB"
        elif self.size < 1000000000:
            return f"{self.size / 1000000.0:#.1f} MB"
        else:
            return f"{self.size / 1000000000.0:#.1f} GB"


class BuildSettings(BaseModel):

    source_path: FilePath
    """Path to the source file for metadata discovery by the parsing plugin."""

    pdf: DownloadableFile
    """The PDF file to display on the landing page."""

    parser: str
    """Name of the parsing plugin."""

    theme: str
    """Name of the theme plugin."""

    canonical_url: Optional[HttpUrl] = None
    """The canonical URL where the landing page is hosted."""

    output_dir: Path = Field(default_factory=lambda: Path("_build"))
    """Path to the output directory for the built site."""

    attachments: List[DownloadableFile] = Field(default_factory=list)
    """List of file attachments to include on the landing page for
    download.
    """

    metadata: Dict[str, Any] = Field(default_factory=dict)
    """Metadata that overrides any metadata discovered by the parsing plugin.
    """

    template_vars: Dict[str, Any] = Field(default_factory=list)
    """Additional variables that are available to the Jinja template
    environment.
    """

    @classmethod
    def load(
        cls,
        *,
        output_dir: Optional[Path] = None,
        source_path: Optional[Path] = None,
        pdf: Optional[Path] = None,
        parser: Optional[str] = None,
        theme: Optional[str] = None,
        canonical_url: Optional[str] = None,
        attachments: Optional[List[Path]] = None,
    ) -> BuildSettings:
        """Create build settings by optionally loadings settings from a
        YAML configuration file and overriding settings from the command line.

        Parameters
        ----------
        output_dir
            Path to the directory where the generate landing page website
            is built (command-line override).
        source_path
            Path to the root source file for metadata extraction by the
            parsing plugin (from the command-line). Typically this is the root
            LaTeX file.
        pdf
            Path to the PDF to display on the landing page (from the
            command-line).
        parser
            Name of the parsing plugin (from the command-line).
        theme
            Name of the theme plugin (from the command-line).
        canonical_url
            The caonical URL where the landing page is hosted.
        attachments
            List of paths to attachments (files) to be included for download
            on the landing page.

        Returns
        -------
        settings
            The ``BuildSettings`` instance.
        """
        settings_path = BuildSettings._get_settings_path(source_path)
        if settings_path:
            settings_data: Dict[str, Any] = yaml.safe_load(
                settings_path.read_text()
            )
        else:
            settings_data = {}

        # Modify the data to convert paths to DownloadableFile tile
        if "pdf" in settings_data:
            settings_data["pdf"] = DownloadableFile.load(
                Path(settings_data["pdf"])
            )
        if "attachments" in settings_data:
            settings_data["attachments"] = [
                DownloadableFile.load(Path(p))
                for p in settings_data["attachments"]
            ]

        # Adding in command-line overrides
        if output_dir:
            settings_data["output_dir"] = output_dir
        if source_path:
            settings_data["source_path"] = source_path
        if pdf:
            settings_data["pdf"] = DownloadableFile.load(pdf)
        if parser:
            settings_data["parser"] = parser
        if theme:
            settings_data["theme"] = theme
        if canonical_url:
            settings_data["canonical_url"] = canonical_url
        if attachments:
            settings_data["attachments"] = [
                DownloadableFile.load(p) for p in attachments
            ]

        return cls.parse_obj(settings_data)

    @staticmethod
    def _get_settings_path(source_path: Optional[Path]) -> Optional[Path]:
        settings_path = Path.cwd().joinpath("lander.yaml")
        if settings_path.exists():
            return settings_path

        if source_path:
            settings_path = source_path.parent.joinpath("lander.yaml")
            if settings_path.exists():
                return settings_path

        return None

    @validator("parser")
    def validate_parser_plugin(cls, v: str) -> str:
        if v not in parsers:
            raise ValueError(
                f"'{v}' is not a known parser plugin. Available parser "
                f"plugins are: {', '.join(parsers.names)}."
            )
        return v

    @validator("theme")
    def validate_theme_plugin(cls, v: str) -> str:
        if v not in themes:
            raise ValueError(
                f"'{v}' is not a known theme plugin. Available themes "
                f"are: {', '.join(themes.names)}."
            )
        return v



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Type

import pkg_resources

if TYPE_CHECKING:
    from lander.ext.parser import Parser


__all__ = ["ParsingPlugins"]


class ParsingPlugins:
    """A class for accessing loadable parsing plugins."""

    def __init__(self, plugins: Dict[str, Type[Parser]]) -> None:
        self.plugins = plugins

    @classmethod
    def load_plugins(cls) -> ParsingPlugins:
        """Load parsing plugins from the ``lander.parser`` setuptools
        entry_point metadata of installed packages.

        Notes
        -----
        Parsing plugins are declared by package's extry_points. This is an
        example fragment from a ``setup.cfg`` file::

        [options.entry_points]
        lander.parsers =
            article = lander.parsers.article:ArticleParser
        """
        discovered_plugins = {
            entry_point.name: entry_point.load()
            for entry_point in pkg_resources.iter_entry_points(
                "lander.parsers"
            )
        }
        return cls(discovered_plugins)

    @property
    def names(self) -> List[str]:
        """The names of available parsing plugins."""
        return sorted(list(self.plugins.keys()))

    def __getitem__(self, key: str) -> Type[Parser]:
        """Get the plugin for the given name."""
        return self.plugins[key]

    def __contains__(self, key: str) -> bool:
        """Determine if the plugins is available, by name."""
        return key in self.plugins



"""Functions for normalizing TeX source."""

from __future__ import annotations

import logging
import os
import re
from pathlib import Path
from typing import Dict, Optional

__all__ = [
    "remove_comments",
    "remove_trailing_whitespace",
    "read_tex_file",
    "process_inputs",
    "replace_macros",
]


# Regular expression for finding input or include commands
input_include_pattern = re.compile(
    r"\\(?P<command>input|include)"  # command name
    r"(([ ]*\{)|([ ]+))"
    r"(?P<filename>[\w/\-\.]+)"  # included filename
    r"[\}%\s]"  # closing whitespace or bracket
)
"""Regular expression for finding input or include commands."""

input_pattern = re.compile(r"\\input[ ]*?{*?(?P<filename>[\w/\-\.]+)[\}%\s]")
"""Regular expression for finding ``input`` commands."""

include_pattern = re.compile(
    r"\\include[ ]*?{*?(?P<filename>[\w/\-\.]+)[\}%\s]"
)
"""Regular expression for finding ``include`` commands."""


def read_tex_file(root_filepath: Path, root_dir: Optional[Path] = None) -> str:
    r"""Read a TeX file, automatically processing and normalizing it
    (including other input files, removing comments, and deleting trailing
    whitespace).

    Parameters
    ----------
    root_filepath
        Filepath to a TeX file.
    root_dir
        Root directory of the TeX project. This only needs to be set when
        recursively reading in ``\input`` or ``\include`` files and the root
        directory is not the directory of ``root_filepath``.

    Returns
    -------
    tex_source
        TeX source.

    Notes
    -----
    This function runs the following pipeline:

    1. `remove_comments`
    2. `remove_trailing_whitespace`
    3. `process_inputs`
    """
    if not root_filepath.is_file:
        raise ValueError(
            f"root_filepath must be a file (got {root_filepath})."
        )
    tex_source = root_filepath.read_text()

    if root_dir is None:
        root_dir = root_filepath.parent

    # Text processing pipline
    tex_source = remove_comments(tex_source)
    tex_source = remove_trailing_whitespace(tex_source)
    tex_source = process_inputs(tex_source, root_dir=root_dir)

    return tex_source


def remove_comments(tex_source: str) -> str:
    """Delete latex comments from TeX source.

    Parameters
    ----------
    tex_source
        TeX source content.

    Returns
    -------
    tex_source
        TeX source without comments.
    """
    # Expression via http://stackoverflow.com/a/13365453
    return re.sub(r"(?<!\\)%.*$", r"", tex_source, flags=re.M)


def remove_trailing_whitespace(tex_source: str) -> str:
    """Delete trailing whitespace from TeX source.

    Parameters
    ----------
    tex_source
        TeX source content.

    Returns
    -------
    tex_source
        TeX source without trailing whitespace.
    """
    # Delete any space or tab characters right before a new line
    return re.sub(r"[ \t]+$", "", tex_source, flags=re.M)


def process_inputs(tex_source: str, root_dir: Optional[Path] = None) -> str:
    r"""Insert referenced TeX file contents (from  ``\input`` and ``\include``
    commands) into the source.

    Parameters
    ----------
    tex_source
        TeX source where referenced source files will be found and inserted.
    root_dir
        Name of the directory containing the TeX project's root file. Files
        referenced by TeX ``\input`` and ``\include`` commands are relative to
        this directory. If not set, the current working directory is assumed.

    Returns
    -------
    tex_source
        TeX source.

    See also
    --------
    `read_tex_file`
        Recommended API for reading a root TeX source file and inserting
        referenced files.
    """
    logger = logging.getLogger(__name__)

    def _sub_line(match: re.Match) -> str:
        """Function to be used with re.sub to inline files for each match."""
        fname = match.group("filename")
        if not fname.endswith(".tex"):
            full_fname = ".".join((fname, "tex"))
        else:
            full_fname = fname
        full_path = os.path.abspath(os.path.join(str(root_dir), full_fname))

        try:
            included_source = read_tex_file(Path(full_path), root_dir=root_dir)
        except IOError:
            logger.error("Cannot open {0} for inclusion".format(full_path))
            raise
        else:
            return included_source

    tex_source = input_include_pattern.sub(_sub_line, tex_source)
    return tex_source


def replace_macros(tex_source: str, macros: Dict[str, str]) -> str:
    r"""Replace macros in the TeX source with their content.

    Parameters
    ----------
    tex_source
        TeX source content.
    macros
        Keys are macro names (including leading ``\``) and values are the
        content (as `str`) of the macros.

    Returns
    -------
    tex_source : `str`
        TeX source with known macros replaced.

    Notes
    -----
    Macros with arguments are not supported.

    Examples
    --------
    >>> macros = {r'\handle': 'LDM-nnn'}
    >>> sample = r'This is document \handle.'
    >>> replace_macros(sample, macros)
    'This is document LDM-nnn.'

    Any trailing slash after the macro command is also replaced by this
    function.

    >>> macros = {r'\product': 'Data Management'}
    >>> sample = r'\title    [Test Plan]  { \product\ Test Plan}'
    >>> replace_macros(sample, macros)
    '\\title    [Test Plan]  { Data Management Test Plan}'
    """
    for macro_name, macro_content in macros.items():
        # '\\?' suffix matches an optional trailing '\' that might be used
        # for spacing.
        pattern = re.escape(macro_name) + r"\\?"
        # Wrap macro_content in lambda to avoid processing escapes
        tex_source = re.sub(pattern, lambda _: macro_content, tex_source)
    return tex_source



__all__ = ["ThemePluginDirectory", "ThemePlugin", "ThemeTemplateLoader"]

from lander.ext.theme._base import ThemePlugin
from lander.ext.theme._discovery import ThemePluginDirectory
from lander.ext.theme._jinjaloader import ThemeTemplateLoader



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Type

import pkg_resources

if TYPE_CHECKING:
    from lander.ext.theme import ThemePlugin

__all__ = ["ThemePluginDirectory"]


class ThemePluginDirectory:
    """A class for accessing theme plugins."""

    def __init__(self, plugins: Dict[str, Type[ThemePlugin]]) -> None:
        self.plugins = plugins

    @classmethod
    def load_plugins(cls) -> ThemePluginDirectory:
        """Load theme plugins from the ``lander.themes`` setuptools
        entry_point metadata.
        """
        discovered_plugins = {
            entry_point.name: entry_point.load()
            for entry_point in pkg_resources.iter_entry_points("lander.themes")
        }
        return cls(discovered_plugins)

    @property
    def names(self) -> List[str]:
        """The names of available parsing plugins."""
        return sorted(list(self.plugins.keys()))

    def __getitem__(self, key: str) -> Type[ThemePlugin]:
        """Get the plugin for the given name."""
        return self.plugins[key]

    def __contains__(self, key: str) -> bool:
        """Determine if the plugins is available, by name."""
        return key in self.plugins



"""Support the Lander parsing extensions."""

__all__ = [
    "DocumentMetadata",
    "Parser",
    "Contributor",
    "Person",
    "FormattedString",
    "ParsingPlugins",
    "GitFile",
    "GitRepository",
    "GitRefType",
    "CiPlatform",
    "CiMetadata",
]

from lander.ext.parser._cidata import CiMetadata, CiPlatform, GitRefType
from lander.ext.parser._datamodel import (
    Contributor,
    DocumentMetadata,
    FormattedString,
    Person,
)
from lander.ext.parser._discovery import ParsingPlugins
from lander.ext.parser._gitdata import GitFile, GitRepository
from lander.ext.parser._parser import Parser



from __future__ import annotations

from pathlib import Path

from lander.ext.theme import ThemePlugin

__all__ = ["BaseTheme"]


class BaseTheme(ThemePlugin):
    """The base landing page theme that other themes are built on."""

    @property
    def name(self) -> str:
        """Name of this theme."""
        return "base"

    @property
    def site_dir(self) -> Path:
        """Site directory containing assets and stub files to render."""
        return Path(__file__).parent.joinpath("site")

    @property
    def templates_dir(self) -> Path:
        """Directory containing Jinja templates included by Jinja-templated
        files in the site.
        """
        return Path(__file__).parent.joinpath("templates")



"""Jinja2 filters."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

import jinja2

if TYPE_CHECKING:
    import datetime

__all__ = ["filter_simple_date", "filter_paragraphify"]


def filter_simple_date(value: datetime.datetime) -> str:
    """Filter a `datetime.datetime` into a 'YYYY-MM-DD' string."""
    return value.strftime("%Y-%m-%d")


def filter_paragraphify(value: str) -> str:
    """Convert text into one or more paragraphs, including <p> tags.

    Based on https://gist.github.com/cemk/1324543
    """
    value = re.sub(r"\r\n|\r|\n", "\n", value)  # Normalize newlines
    paras = re.split("\n{2,}", value)
    paras = ["<p>{0}</p>".format(p) for p in paras if len(p) > 0]
    return jinja2.Markup("\n\n".join(paras))



"""Interface to SPDX license metadata."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List

from pydantic import BaseModel, Field, HttpUrl

__all__ = ["SpdxLicense", "SpdxFile", "Licenses"]


class SpdxLicense(BaseModel):
    """A license item as contained in a SpdxFile."""

    name: str
    """Common human-readable name of the license."""

    license_id: str = Field(alias="licenseId")
    """SPDX license identifier."""

    see_also: List[HttpUrl] = Field(alias="seeAlso")
    """"URLs to documentation about this license."""

    is_osi_approved: bool = Field(alias="isOsiApproved")
    """Flag indicating if the license is approved by the OSI
    (Open Software Initiative)
    """


class SpdxFile(BaseModel):
    """Representation of a SPDX license database file as a Pydantic model."""

    licenseListVersion: str
    """Version string of this license file."""

    licenses: List[SpdxLicense]
    """The licenses."""

    @classmethod
    def load_internal(cls) -> SpdxFile:
        """Load an SpdxFile from Lander's package data.

        Returns
        -------
        spdx_file: `SpdxFile`
            The SpdxFile instance from the licenses.json file packaged in
            Lander. This file is a mirror of the canonical license database
            maintained at
            https://github.com/spdx/license-list-data/blob/master/json/licenses.json.
        """
        p = Path(__file__).parent / "data" / "licenses.json"
        return cls.parse_file(p, content_type="application/json")


@dataclass
class Licenses:
    """License database, with access by license ID."""

    licenses: Dict[str, SpdxLicense]
    """Internal license dictionary."""

    @classmethod
    def load(cls) -> Licenses:
        """Load licenses from the SPDX database packaged in Lander.

        Returns
        -------
        licenses : `Licenses`
            The license database instance.
        """
        spdx_file = SpdxFile.load_internal()
        licenses: Dict[str, SpdxLicense] = {
            license.license_id: license for license in spdx_file.licenses
        }
        return cls(licenses=licenses)

    def __getitem__(self, spdx_id: str) -> SpdxLicense:
        return self.licenses[spdx_id]

    def __contains__(self, spdx_id: str) -> bool:
        return spdx_id in self.licenses



"""HTML renderering functions."""

__all__ = [
    "create_jinja_env",
    "render_homepage",
    "filter_simple_date",
    "filter_paragraphify",
]

import datetime
import os
import re
from typing import TYPE_CHECKING

import jinja2

if TYPE_CHECKING:
    from lander.config import Configuration


def create_jinja_env() -> jinja2.Environment:
    """Create a Jinja2 `~jinja2.Environment`.

    Returns
    -------
    env : `jinja2.Environment`
        Jinja2 template rendering environment, configured to use templates in
        ``templates/``.
    """
    template_dir = os.path.join(os.path.dirname(__file__), "templates")
    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(template_dir),
        autoescape=jinja2.select_autoescape(["html"]),
    )
    env.filters["simple_date"] = filter_simple_date
    env.filters["paragraphify"] = filter_paragraphify
    return env


def render_homepage(config: "Configuration", env: jinja2.Environment) -> str:
    """Render the homepage.jinja template."""
    template = env.get_template("homepage.jinja")
    rendered_page = template.render(config=config)
    return rendered_page


def filter_simple_date(value: datetime.datetime) -> str:
    """Filter a `datetime.datetime` into a 'YYYY-MM-DD' string."""
    return value.strftime("%Y-%m-%d")


def filter_paragraphify(value: str) -> str:
    """Convert text into one or more paragraphs, including <p> tags.

    Based on https://gist.github.com/cemk/1324543
    """
    value = re.sub(r"\r\n|\r|\n", "\n", value)  # Normalize newlines
    paras = re.split("\n{2,}", value)
    paras = ["<p>{0}</p>".format(p) for p in paras if len(p) > 0]
    return jinja2.Markup("\n\n".join(paras))



"""Jinja template loader for theme plugins."""

from __future__ import annotations

import logging
import re
from os.path import getmtime
from typing import TYPE_CHECKING, Callable, Optional, Tuple

from jinja2 import BaseLoader, TemplateNotFound

if TYPE_CHECKING:
    from pathlib import Path

    from jinja2 import Environment

    from lander.ext.theme._base import ThemePlugin


__all__ = ["ThemeTemplateLoader"]


SITE_FILE_PATTERN = re.compile(r"^\$(?P<theme>[a-zA-Z0-9]+)\/(?P<path>.+)$")


class ThemeTemplateLoader(BaseLoader):
    """Jinja2 Template loader for theme templates.

    Parameters
    ----------
    theme : `lander.ext.theme.ThemePlugin`
        The theme plugin associated with this template loader.

    Notes
    -----
    The template loader deals with two types of templates:

    - Site file templates. These templates are located with the "site"
      directory (`lander.ext.theme.ThemePlugin.site_dir`) of a theme and
      correspond to pages in the landing page. When referencing these
      templates using the `get_source` method, prefix the path of the template
      with this pattern: ``$<theme name>/<template path>``. For example:
      ``$base/index.html.jinja``.

    - Templates indended to be included by site file templates (using the
      Jinja2 ``include`` syntax). These templates are simply referenced by
      their relative path within a theme's templates directory
      (`lander.ext.theme.ThemePlugin.templates_dir`)

    In addition to these theme template types, this template loader also
    implements template inheritance. If a template isn't found in a theme,
    the loader calls the loader of the base theme (set via the theme's
    `lander.ext.theme.ThemePlugin.base_theme_name` property).
    """

    def __init__(self, theme: ThemePlugin) -> None:
        self._logger = logging.getLogger(__name__)
        self._theme = theme

    @property
    def theme(self) -> ThemePlugin:
        """The theme plugin."""
        return self._theme

    @property
    def inherited_loader(self) -> Optional[ThemeTemplateLoader]:
        """The template loader from the theme's base theme (if the theme
        inherits from a base theme).
        """
        if self._theme.base_theme is not None:
            return self._theme.base_theme.template_loader
        else:
            return None

    def get_source(
        self, environment: Environment, template: str
    ) -> Tuple[str, str, Callable]:
        """Get the template source given its name

        Implements the Jinja `jinja.BaseLoader` interface.
        """
        if template.startswith("$"):
            match = SITE_FILE_PATTERN.match(template)
            if match:
                theme_name = match["theme"]
                file_path = match["path"]
                return self.get_site_file_template(
                    environment, theme_name, file_path
                )
            else:
                raise TemplateNotFound(template)
        else:
            return self.get_theme_template(environment, template)

    def get_site_file_template(
        self, environment: Environment, theme_name: str, file_path: str
    ) -> Tuple[str, str, Callable]:
        """Get the source for a templated site file."""
        if theme_name == self.theme.name:
            path = self.theme.site_dir.joinpath(file_path)
            self._logger.debug(
                "%s loader: looking for site file at path %s",
                self.theme.name,
                path,
            )
            if not path.is_file():
                raise TemplateNotFound(file_path)
            return self._return_source(path)
        elif self.inherited_loader is not None:
            return self.inherited_loader.get_site_file_template(
                environment, theme_name, file_path
            )
        else:
            self._logger.debug(
                "%s loader: no inherited loaders available for theme name %s",
                self.theme.name,
                theme_name,
            )
            raise TemplateNotFound(file_path)

    def get_theme_template(
        self, environment: Environment, template: str
    ) -> Tuple[str, str, Callable]:
        """Get the source for a theme template, or a template in an inherited
        theme.
        """
        path = self.theme.templates_dir.joinpath(template)
        if path.is_file():
            return self._return_source(path)
        elif self.inherited_loader is not None:
            return self.inherited_loader.get_theme_template(
                environment, template
            )
        else:
            raise TemplateNotFound(template)

    def _return_source(self, path: Path) -> Tuple[str, str, Callable]:
        """Return template source following the `Loader.get_source`
        return-type specification.
        """
        source = path.read_text()
        modified_time = getmtime(path)
        return (
            source,
            str(path.resolve()),
            lambda: modified_time == getmtime(path),
        )



"""Access loadable plugins."""

from __future__ import annotations

from lander.ext.parser import ParsingPlugins
from lander.ext.theme import ThemePluginDirectory

__all__ = ["parsers", "themes"]


parsers = ParsingPlugins.load_plugins()
"""Parsing plugins directory."""

themes = ThemePluginDirectory.load_plugins()
"""Theme plugins directory."""



"""Metadata from the continuous integration environment."""

from __future__ import annotations

import os
import re
from dataclasses import dataclass
from enum import Enum
from typing import Optional

__all__ = ["GitRefType", "CiMetadata", "GitRefType"]


class GitRefType(str, Enum):
    """Type of git reftag, or branch."""

    tag = "tag"
    """The git ref corresponds to a tag."""

    branch = "branch"
    """The git ref corresponds to a branch."""


class CiPlatform(str, Enum):
    """The name of the CI platform."""

    null = "null"
    """The build is not performed through a common CI platform."""

    github_actions = "github_actions"
    """The build is performed on GitHub Actions."""

    travis = "travis"
    """The build is performed on Travis CI."""


@dataclass
class CiMetadata:
    """Metadata gathered from CI platform environment variables."""

    platform: CiPlatform = CiPlatform.null
    """The CI platform."""

    git_ref: Optional[str] = None
    """Git reference: branch or tag name."""

    git_ref_type: Optional[GitRefType] = None
    """Type of git reference: branch or tag."""

    git_sha: Optional[str] = None
    """The SHA1 of the Git commit."""

    build_id: Optional[str] = None
    """CI build number."""

    build_url: Optional[str] = None
    """CI dashboard URL for this build."""

    github_slug: Optional[str] = None
    """Slug of the GitHub repository (``org/repo``)."""

    @property
    def github_repository(self) -> Optional[str]:
        """URL of the GitHub repository homepage."""
        if self.github_slug:
            return f"https://github.com/{self.github_slug}"
        else:
            return None

    @classmethod
    def create(cls) -> CiMetadata:
        """Gather CI metadata, automatically inferring the CI platform."""
        if os.getenv("GITHUB_ACTIONS"):
            return cls.for_github_actions()
        elif os.getenv("TRAVIS") == "true":
            return cls.for_travis()
        else:
            # CI platform cannot be inferred, so return empty metadata.
            return cls()

    @classmethod
    def for_github_actions(cls) -> CiMetadata:
        """Gather CI metadata from a GitHub Actions environment."""
        github_ref = os.getenv("GITHUB_REF")
        run_id = os.getenv("GITHUB_RUN_ID")
        repo = os.getenv("GITHUB_REPOSITORY")

        if run_id and repo:
            ci_build_url: Optional[str] = (
                f"https://github.com/{repo}/actions/runs/{run_id}"
            )
        else:
            ci_build_url = None

        if github_ref:
            m = re.match(
                r"refs/(?P<kind>heads|tags|pull)/(?P<ref>.+)", github_ref
            )
            if m:
                if m.group("kind") == "heads":
                    git_ref_type: Optional[GitRefType] = GitRefType.branch
                elif m.group("kind") == "tags":
                    git_ref_type = GitRefType.tag
                else:
                    git_ref_type = None

                git_ref: Optional[str] = m.group("ref")
            else:
                git_ref_type = None
                git_ref = None

        return cls(
            platform=CiPlatform.github_actions,
            git_sha=os.getenv("GITHUB_SHA"),
            git_ref=git_ref,
            git_ref_type=git_ref_type,
            build_id=run_id,
            build_url=ci_build_url,
            github_slug=repo,
        )

    @classmethod
    def for_travis(cls) -> CiMetadata:
        """Gather CI metadata from a Travis environment."""
        if os.getenv("TRAVIS_TAG"):
            git_ref: Optional[str] = os.getenv("TRAVIS_TAG")
            git_ref_type: Optional[GitRefType] = GitRefType.tag
        elif os.getenv("TRAVIS_BRANCH"):
            git_ref = os.getenv("TRAVIS_BRANCH")
            git_ref_type = GitRefType.branch
        else:
            git_ref = None
            git_ref_type = None

        return cls(
            platform=CiPlatform.travis,
            git_sha=os.getenv("TRAVIS_COMMIT"),
            git_ref=git_ref,
            git_ref_type=git_ref_type,
            build_id=os.getenv("TRAVIS_BUILD_NUMBER"),
            build_url=os.getenv("TRAVIS_BUILD_WEB_URL"),
            github_slug=os.getenv("TRAVIS_REPO_SLUG"),
        )



from __future__ import annotations

import datetime
import re
from typing import TYPE_CHECKING, Any, Generator, List, Optional

import base32_lib as base32
import bleach
from pydantic import BaseConfig, BaseModel, EmailStr, Field, HttpUrl, validator
from pydantic.errors import UrlError
from pydantic.fields import ModelField

from lander.ext.parser.pandoc import convert_text
from lander.spdx import Licenses

if TYPE_CHECKING:
    from pydantic.typing import AnyCallable

    CallableGenerator = Generator[AnyCallable, None, None]


__all__ = [
    "FormattedString",
    "Person",
    "Contributor",
    "Orcid",
    "DocumentMetadata",
]

WHITESPACE_PATTERN = re.compile(r"\s+")

ORCID_PATTERN = re.compile(
    r"(?P<identifier>[0-9X]{4}-[0-9X]{4}-[0-9X]{4}-[0-9X]{4})"
)
"""Regular expression for matching the ORCiD identifier

Examples:

- 0000-0002-1825-0097
- 0000-0001-5109-3700
- 0000-0002-1694-233X

For more information, see
https://support.orcid.org/hc/en-us/articles/360006897674
"""

ROR_PATTERN = re.compile(
    r"https://ror.org"
    r"\/(?P<identifier>0[0-9abcdefghjkmpqrstuvwxyzabcdefghjkmpqrstuvwxyz]{8})",
    flags=re.IGNORECASE,
)


def collapse_whitespace(text: str) -> str:
    """Replace any whitespace character, or group, with a single space."""
    return WHITESPACE_PATTERN.sub(" ", text).strip()


class FormattedString(BaseModel):
    """A string with plain and HTML encodings."""

    html: str
    """HTML version of the string."""

    plain: str
    """Plain (unicode) version of the string."""

    latex: Optional[str]
    """LaTeX version of the string, if available."""

    @classmethod
    def from_latex(cls, tex: str, fragment: bool = False) -> FormattedString:
        """Created a FormattedString from LaTeX-encoded content.

        The LaTeX content is transformed to HTML and plain encodings
        via pandoc.
        """
        try:
            plain = convert_text(
                content=tex,
                source_fmt="latex",
                output_fmt="plain",
                deparagraph=fragment,
            )
        except Exception:
            plain = tex  # fallback in case conversion didn't work

        try:
            html = convert_text(
                content=tex,
                source_fmt="latex",
                output_fmt="html",
                deparagraph=fragment,
            )
        except Exception:
            html = tex  # fallback in case conversion didn't work

        return cls(html=html, plain=plain, latex=tex)

    @validator("html")
    def santize_html(cls, v: str) -> str:
        """Ensure that the HTML is safe for injecting into templates."""
        # Add <p> to the default list of allowed tags, which is useful for
        # abstracts.
        return bleach.clean(
            v,
            strip=True,
            tags=[
                "p",
                "a",
                "abbr",
                "acronym",
                "b",
                "blockquote",
                "code",
                "em",
                "i",
                "li",
                "ol",
                "strong",
                "ul",
            ],
        )

    @validator("*")
    def clean_whitespace(cls, v: str) -> str:
        return collapse_whitespace(v)


class OrcidError(UrlError):
    code = "orcid"
    msg_template = "invalid ORCiD"


class Orcid(HttpUrl):
    """An ORCiD type for Pydantic validation.

    The validator forces an ORCiD identifier to always be a URL for
    orcid.org, per https://support.orcid.org/hc/en-us/articles/360006897674.
    This validator implments the ISO 7064 11,2 checksum algorithm.
    """

    allowed_schemes = {"https"}

    @classmethod
    def __get_validators__(cls) -> "CallableGenerator":
        yield cls.validate

    @classmethod
    def validate(
        cls, value: Any, field: ModelField, config: BaseConfig
    ) -> Orcid:
        if value.__class__ == cls:
            return value

        m = ORCID_PATTERN.search(value)
        if not m:
            raise OrcidError()

        identifier = m["identifier"]
        if not cls.verify_checksum(identifier):
            raise OrcidError()

        return HttpUrl.validate(
            f"https://orcid.org/{identifier}", field, config
        )

    @staticmethod
    def verify_checksum(identifier: str) -> bool:
        """Verify the checksum of an ORCiD identifier string (path component
        of the URL) given teh ISO 7064 11,2 algorithm.
        """
        total: int = 0
        for digit in identifier:
            if digit == "X":
                digit = "10"
            if not digit.isdigit():
                continue
            total = (total + int(digit)) * 2
        remainder = total % 11
        result = (12 - remainder) % 11
        return result == 10


class RorError(UrlError):
    code = "ror"
    msg_template = "invalid ROR"


class Ror(HttpUrl):
    """A ROR (Research Organization Registry) type for Pydantic validation.
    """

    allowed_schemes = {"https"}

    @classmethod
    def __get_validators__(cls) -> "CallableGenerator":
        yield cls.validate

    @classmethod
    def validate(
        cls, value: Any, field: ModelField, config: BaseConfig
    ) -> Ror:
        if value.__class__ == cls:
            return value

        m = ROR_PATTERN.search(value)
        if not m:
            print("pattern did not match")
            raise RorError()

        identifier = m["identifier"]
        try:
            base32.decode(identifier, checksum=True)
        except ValueError:
            raise RorError()

        return HttpUrl.validate(value, field, config)


class Organization(BaseModel):
    """Data about an organization (often used as an affiliation).
    """

    name: str
    """The display name of the institution."""

    ror: Optional[Ror] = None
    """The ROR identifier of the institution."""

    address: Optional[str] = None
    """The address of the institution."""

    url: Optional[HttpUrl] = None
    """The homepage of the institution."""


class Person(BaseModel):
    """Data about a person."""

    name: str
    """Display name of the person."""

    orcid: Optional[Orcid] = None
    """The ORCiD of the person."""

    affiliations: Optional[List[Organization]] = Field(
        default_factory=lambda: []
    )
    """The person's affiliations."""

    email: Optional[EmailStr] = None
    """Email associated with the person."""

    @validator("name")
    def clean_whitespace(cls, v: str) -> str:
        return collapse_whitespace(v)


class Contributor(Person):
    """Data about a contributor.

    A ``Contributor`` is the same as a ``Person``, with the addition of the
    `role` attribute.
    """

    role: Optional[str] = None
    """Description of the contributor's role."""


class DocumentMetadata(BaseModel):
    """A container for LaTeX document metadata."""

    title: str
    """Document title."""

    identifier: Optional[str] = None
    """Document identifier."""

    abstract: Optional[FormattedString] = None
    """Document abstract or summary."""

    authors: List[Contributor] = Field(default_factory=lambda: [])
    """Authors of the document."""

    date_modified: Optional[datetime.date] = None
    """Date when the document was last modified."""

    version: Optional[str] = None
    """Version of this document."""

    keywords: List[str] = Field(default_factory=lambda: [])
    """Keywords associated with the document."""

    repository_url: Optional[HttpUrl]
    """URL of the document's source code repository."""

    copyright: Optional[str]
    """Free-form copyright statement."""

    license_identifier: Optional[str]
    """The license of the document, as an SPDX identifier.

    See https://spdx.org/licenses/ for a list of licenses.
    """

    full_text: Optional[FormattedString] = None
    """The full text content document."""

    ci_url: Optional[HttpUrl] = None
    """The URL of the continuous integration build for the document."""

    canonical_url: Optional[HttpUrl] = None
    """The canonical URL where this document is published."""

    @validator("title", "version", "keywords", "copyright", each_item=True)
    def clean_whitespace(cls, v: str) -> str:
        return collapse_whitespace(v)

    @validator("license_identifier")
    def validate_spdx(cls, v: Optional[str]) -> Optional[str]:
        if v is not None:
            licenses = Licenses.load()
            if v not in licenses:
                raise ValueError(
                    f"License ID '{v}' is not a valid SPDX license identifier."
                )
        return v

    def get_license_name(self) -> Optional[str]:
        """Get the name of the license."""
        if self.license_identifier is not None:
            licenses = Licenses.load()
            license = licenses[self.license_identifier]
            return license.name
        else:
            return None



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Dict, Optional

from lander.ext.parser._cidata import CiMetadata
from lander.ext.parser._datamodel import DocumentMetadata
from lander.ext.parser._gitdata import GitRepository
from lander.ext.parser.texutils.extract import get_macros
from lander.ext.parser.texutils.normalize import read_tex_file, replace_macros

if TYPE_CHECKING:
    from pathlib import Path
    from lander.settings import BuildSettings


__all__ = ["Parser"]


class Parser(metaclass=ABCMeta):
    """Base class for TeX document metadata parsing extensions.

    Parameters
    ----------
    settings : `lander.settings.BuildSettings`
        The build settings for this site, which includes command-line and YAML
        configuration overrides of metadata.
    """

    def __init__(self, *, settings: BuildSettings) -> None:
        self._settings = settings

        _tex_source = read_tex_file(self.tex_path)
        self._tex_macros = get_macros(_tex_source)
        self._tex_source = self.normalize_source(_tex_source)

        try:
            self._git_repository: Optional[
                GitRepository
            ] = GitRepository.create(self.tex_path.parent)
        except Exception:
            self._git_repository = None

        self._ci_metadata = CiMetadata.create()

        self._metadata = self.extract_metadata()

    @property
    def settings(self) -> BuildSettings:
        """The build settings."""
        return self._settings

    @property
    def tex_path(self) -> Path:
        """"Path to the root TeX source file."""
        return self.settings.source_path

    @property
    def tex_source(self) -> str:
        """TeX source, which has been normalized."""
        return self._tex_source

    @property
    def tex_macros(self) -> Dict[str, str]:
        """TeX macros detected by
        `lander.ext.parser.texutils.extract.get_macros`.

        Keys are command names (including the slash) and values are the values
        of those macros.

        This property is useful because the normalized source in `tex_source`
        typically has macro definitions clobbered.
        """
        return self._tex_macros

    @property
    def ci_metadata(self) -> CiMetadata:
        """Metadata from the CI environment

        This attribute is instantiate automatically and is available to the
        `extract_metadata` hook for use by parser implementations.
        """
        return self._ci_metadata

    @property
    def git_repository(self) -> Optional[GitRepository]:
        """Metadata from the local Git repository

        This attribute is instantiate automatically and is available to the
        `extract_metadata` hook for use by parser implementations.
        """
        return self._git_repository

    @property
    def metadata(self) -> DocumentMetadata:
        """Metadata about the document."""
        return self._metadata

    def normalize_source(self, tex_source: str) -> str:
        """Process the TeX source after it is read, but before metadata
        is extracted.

        Parameters
        ----------
        tex_source
            TeX source content.

        Returns
        -------
        tex_source
            Normalized TeX source content.
        """
        macros = get_macros(tex_source)
        return replace_macros(tex_source, macros)

    @abstractmethod
    def extract_metadata(self) -> DocumentMetadata:
        """Hook for implementing metadata extraction.

        Returns
        -------
        metadata
            The metadata parsed from the document source.
        """
        raise NotImplementedError



"""Pandoc-based markup conversion APIs."""

from __future__ import annotations

import functools
import logging
from typing import Any, Callable, List, Optional, TypeVar

import pypandoc

__all__ = ["convert_text", "ensure_pandoc"]

F = TypeVar("F", bound=Callable[..., Any])


def ensure_pandoc(func: F) -> Callable[..., Any]:
    """Decorate a function that uses pypandoc to ensure that pandoc is
    installed if necessary.
    """
    logger = logging.getLogger(__name__)

    @functools.wraps(func)
    def _install_and_run(*args: Any, **kwargs: Any) -> Any:
        try:
            # First try to run pypandoc function
            result = func(*args, **kwargs)
        except OSError:
            # Install pandoc and retry
            message = "Pandoc is required but not found."
            logger.warning(message)

            result = func(*args, **kwargs)

        return result

    return _install_and_run


@ensure_pandoc
def convert_text(
    *,
    content: str,
    source_fmt: str,
    output_fmt: str,
    deparagraph: bool = False,
    mathjax: bool = False,
    extra_args: Optional[List[str]] = None,
) -> str:
    """Convert text from one markup format to another using pandoc.

    This function is a thin wrapper around `pypandoc.convert_text`.

    Parameters
    ----------
    content : `str`
        Original content.

    source_fmt : `str`
        Format of the original ``content``. Format identifier must be one of
        those known by Pandoc. See https://pandoc.org/MANUAL.html for details.

    output_fmt : `str`
        Output format for the content.

    deparagraph : `bool`, optional
        If `True`, then the
        `lander.parse.pandoc.filters.deparagraph.deparagraph` filter is
        used to remove paragraph (``<p>``, for example) tags around a single
        paragraph of content. That filter does not affect content that
        consists of multiple blocks (several paragraphs, or lists, for
        example). Default is `False`.

        For example, **without** this filter Pandoc will convert
        the string ``"Title text"`` to ``"<p>Title text</p>"`` in HTML. The
        paragraph tags aren't useful if you intend to wrap the converted
        content in different tags, like ``<h1>``, using your own templating
        system.

        **With** this filter, Pandoc will convert the string ``"Title text"``
        to ``"Title text"`` in HTML.

    mathjax : `bool`, optional
        If `True` then Pandoc will markup output content to work with MathJax.
        Default is False.

    extra_args : `list`, optional
        Sequence of Pandoc arguments command line arguments (such as
        ``'--normalize'``). The ``deparagraph``, and ``mathjax``
        arguments are convenience arguments that are equivalent to items
        in ``extra_args``.

    Returns
    -------
    output : `str`
        Content in the output (``to_fmt``) format.

    Notes
    -----
    This function will automatically install Pandoc if it is not available.
    See `ensure_pandoc`.
    """
    logger = logging.getLogger(__name__)

    if extra_args is not None:
        extra_args = list(extra_args)
    else:
        extra_args = []

    if mathjax:
        extra_args.append("--mathjax")

    if deparagraph:
        extra_args.append("--filter=lander-deparagraph")

    extra_args.append("--wrap=none")

    # de-dupe extra args
    extra_args = list(set(extra_args))

    logger.debug(
        "Running pandoc from %s to %s with extra_args %s",
        source_fmt,
        output_fmt,
        extra_args,
    )

    output = pypandoc.convert_text(
        content, output_fmt, format=source_fmt, extra_args=extra_args
    )
    return output



"""This modules deals with generating metadata.jsonld files from parsed
document metadata.

This metadata format is considered legacy for compatibility with the orignal
versions of Lander (versions 0.x).
"""

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, List, Optional

from pydantic import BaseModel, Field, HttpUrl

__all__ = ["CodemetaPerson", "CodemetaData"]

if TYPE_CHECKING:
    from lander.ext.parser import DocumentMetadata


class CodemetaPerson(BaseModel):
    """A JSON-LD presentation of a person (typically an author)."""

    name: str

    at__type: str = Field("Person", alias="@type")


class CodemetaData(BaseModel):
    """A Codemeta/JSON-LD version of document metadata.

    This data model is compatible with the ``metadata.jsonld`` files produced
    by the early versions of Lander (0.x).
    """

    at__id: str = Field("", alias="@id")

    at__context: List[str] = Field(
        alias="@context",
        default_factory=lambda: [
            "https://raw.githubusercontent.com/codemeta/codemeta/2.0-rc/"
            "codemeta.jsonld",
            "http://schema.org",
        ],
    )

    at__type: List[str] = Field(
        alias="@type", default_factory=lambda: ["Report", "SoftwareSourceCode"]
    )

    language: str = "TeX"

    reportNumber: Optional[str] = None

    name: Optional[str] = None

    description: Optional[str] = None

    author: Optional[List[CodemetaPerson]] = Field(default_factory=lambda: [])

    dateModified: Optional[datetime.datetime] = None

    articleBody: Optional[str] = None

    fileFormat: str = "text/plain"

    url: Optional[HttpUrl] = None

    codeRepository: Optional[HttpUrl] = None

    contIntegration: Optional[HttpUrl] = None

    readme: Optional[HttpUrl] = None

    license_id: Optional[str] = None

    @classmethod
    def from_document_metadata(
        cls,
        *,
        metadata: DocumentMetadata,
        url: str,
        ci_url: Optional[str] = None,
    ) -> CodemetaData:
        """Create a CodemetaData from a `lander.ext.parser.DocumentMetadata`
        instance.
        """
        if metadata.full_text:
            full_text: Optional[str] = metadata.full_text.plain
        else:
            full_text = None

        instance = cls(
            at__id=url,
            reportNumber=metadata.identifier,
            name=metadata.title,
            description=metadata.abstract,
            author=[CodemetaPerson(name=a.name) for a in metadata.authors],
            dateModified=metadata.date_modified,
            articleBody=full_text,
            url=url,
            codeRepository=metadata.repository_url,
            contIntegration=metadata.ci_url,
        )
        instance.at__id = url
        return instance

    def jsonld(self) -> str:
        """Export to JSON-LD content."""
        return self.json(by_alias=True)



from __future__ import annotations

import logging
import shutil
from abc import ABCMeta, abstractmethod
from pathlib import Path, PurePath, PurePosixPath
from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional
from urllib.parse import urljoin

import jinja2

from lander.ext.theme._jinjaloader import ThemeTemplateLoader
from lander.ext.theme.jinjafilters import (
    filter_paragraphify,
    filter_simple_date,
)

if TYPE_CHECKING:
    from lander.ext.parser import DocumentMetadata
    from lander.settings import BuildSettings


__all__ = ["ThemePlugin"]


class ThemePlugin(metaclass=ABCMeta):
    """Base class for landing page theme plugins.

    Parameters
    ----------
    metadata : `lander.ext.parser.DocumentMetadata`
        The document's metadata, typically created by the parsing plugin.
    settings : `lander.settings.BuildSettings`
        The build settings for this site, which includes command-line and YAML
        configuration overrides of metadata.
    """

    def __init__(
        self, *, metadata: DocumentMetadata, settings: BuildSettings
    ) -> None:
        self._logger = logging.getLogger(__name__)
        self._base_theme: Optional[ThemePlugin] = None
        self._metadata = metadata
        self._settings = settings

        self._template_loader = self._init_template_loader()
        self._jinja_env = self.create_jinja_env()

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of this theme."""
        raise NotImplementedError

    @property
    def metadata(self) -> DocumentMetadata:
        """The document metadata."""
        return self._metadata

    @property
    def settings(self) -> BuildSettings:
        """The build settings."""
        return self._settings

    @property
    def template_loader(self) -> ThemeTemplateLoader:
        """The Jinja template loader."""
        return self._template_loader

    @property
    def logger(self) -> logging.Logger:
        """Logger for use the theme plugin."""
        return self._logger

    @property
    def jinja_env(self) -> jinja2.Environment:
        """The Jinja environment."""
        return self._jinja_env

    @property
    def base_theme_name(self) -> Optional[str]:
        """Name of the base theme, or `None` if this theme does not inherit
        from another theme.
        """
        return None

    @property
    def base_theme(self) -> Optional[ThemePlugin]:
        """The base theme."""
        if self.base_theme_name is not None:
            if self._base_theme is None:
                from lander.plugins import themes

                self._base_theme = themes[self.base_theme_name](
                    metadata=self.metadata, settings=self.settings
                )
            return self._base_theme
        else:
            return None

    def build_site(self, output_dir: Optional[Path] = None) -> None:
        """Build the landing page site, including rendering templates and
        copying assets into the output directory.

        Parameters
        ----------
        output_dir : `pathlib.Path`, optional
            Directory where the landing page site is built. If the output
            directory is not set, the output directory is determined from
            the `ThemePlugin.settings`.
        """
        if output_dir is None:
            output_dir = self.settings.output_dir
        assert isinstance(output_dir, Path)
        output_dir.mkdir(parents=True, exist_ok=True)

        site_inventory = self._build_site_inventory()
        for relative_path, file_path in site_inventory.items():
            if file_path.suffix == ".jinja":
                self._render_path(file_path, relative_path, output_dir)
            else:
                self._copy_path(file_path, relative_path, output_dir)

        # Copy the PDF
        output_pdf_path = output_dir.joinpath(self.settings.pdf.name)
        shutil.copy(self.settings.pdf.file_path, output_pdf_path)

        # Copy attachments
        for attachment in self.settings.attachments:
            output_attachment_path = output_dir.joinpath(
                self.settings.pdf.name
            )
            shutil.copy(attachment.file_path, output_attachment_path)

        self._write_metadata(output_dir)
        self.run_post_build(output_dir)

    def run_post_build(self, output_dir: Path) -> None:
        """Hook that is executed at the end of the site site build.

        Themes can implement this method to perform tasks on a completed site.

        Parameters
        ----------
        output_dir : `pathlib.Path`
            File system directory where the site is built.
        """
        pass

    def _build_site_inventory(
        self, inventory: Optional[Dict[PurePath, Path]] = None
    ) -> Dict[PurePath, Path]:
        """Create an inventory of files in the site.

        This method gathers file paths in this theme's "site" directory, as
        well as any files in the site directories of base themes. Files in
        a theme override its base theme.

        Returns
        -------
        inventory : `dict`
            Dictionary where the key is the relative path of the file in the
            site. This is a `PurePath`. The value is the filesystem path of
            that file.
        """
        # Map relative site path to filesystem path of the asset.
        if inventory is None:
            inventory = {}
        assert inventory is not None

        # First add paths from the base theme(s)
        if self.base_theme:
            self.base_theme._build_site_inventory(inventory)

        # Add paths from this theme's site directory that add to or override
        # the assets from the base theme
        for path in self._site_dir_contents(self.site_dir):
            path_key = PurePath(path.relative_to(self.site_dir))
            inventory[path_key] = path

        return inventory

    def _site_dir_contents(self, directory: Path) -> Iterator[Path]:
        """Iterate over the contents of a directory, including contents of
        subdirectories.

        Parameters
        ----------
        directory : `pathlib.Path`
            Root directory to find files.

        Yields
        ------
        path : `pathlib.Path`
            Path to a file inside ``directory``.
        """
        for path in directory.iterdir():
            if path.is_dir():
                yield from self._site_dir_contents(path)
            else:
                yield path

    def _copy_path(
        self, site_path: Path, relative_path: PurePath, output_dir: Path
    ) -> None:
        """Copy a path in the theme's site directory into the same relative
        path in the output directory.
        """
        output_path = output_dir.joinpath(relative_path)
        self.logger.debug("Copying %s to %s", relative_path, output_path)
        if not output_path.parent.exists():
            output_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(site_path, output_path)

    def _render_path(
        self, site_path: Path, relative_path: PurePath, output_dir: Path
    ) -> None:
        """Render a Jinja2 template and write it to the same relative path
        in the output directory.

        Notes
        -----
        The output path will be the same as the ``site_path``, but without
        the original ``.jinja`` extension.
        """
        # Relative path of the output (remove the .jinja extension)
        relative_output_path = relative_path.with_suffix("").with_suffix(
            "".join(site_path.suffixes[:-1])
        )
        # Remove the .jinja extension while also locating the rendered output
        # in the build directory.
        output_path = output_dir.joinpath(relative_output_path)
        if not output_path.parent.exists():
            output_path.parent.mkdir(parents=True, exist_ok=True)

        template_name = f"${self.name}/{relative_path!s}"
        self.logger.debug("Rendering templated file: %s", relative_output_path)
        jinja_template = self.jinja_env.get_template(template_name)
        context = self.create_jinja_context(
            path=PurePosixPath(relative_output_path),
            template_name=template_name,
        )
        content = jinja_template.render(**context)

        output_path.write_text(content)

    def _init_template_loader(self) -> ThemeTemplateLoader:
        """Initialize the custom Jinja template loader."""
        return ThemeTemplateLoader(self)

    def create_jinja_env(self) -> jinja2.Environment:
        """Create a Jinja environment, with the template loader and filters
        set.
        """
        env = jinja2.Environment(
            loader=self.template_loader,
            autoescape=jinja2.select_autoescape(["html"]),
        )
        env.filters["simple_date"] = filter_simple_date
        env.filters["paragraphify"] = filter_paragraphify
        return env

    def create_jinja_context(
        self, *, path: PurePosixPath, template_name: str
    ) -> Dict[str, Any]:
        """Create the context for rendering a Jinja template.

        This method can be implemented by themes to customize the Jinja
        context.

        Parameters
        ----------
        path : `pathlib.PurePosixPath`
            The relative path of the templated context in the site. This path
            is used to set the ``canonical_url`` template variable.
        template_name: `str`
            Name of the Jinja template.

        Returns
        -------
        context : `dict`
            Dictionary used as the Jinja template rendering context. Keys
            are variable names in templates.
        """
        context: Dict[str, Any] = {
            "metadata": self.metadata,
            "settings": self.settings.template_vars,
            "pdf": self.settings.pdf,
            "attachments": self.settings.attachments,
        }
        if self.settings.canonical_url:
            context["canonical_url"] = urljoin(
                self.settings.canonical_url, str(path)
            )
            # Strip "index.html" from the canonical URL
            if context["canonical_url"].endswith("/index.html"):
                context["canonical_url"] = context["canonical_url"][:-10]
        return context

    @property
    @abstractmethod
    def site_dir(self) -> Path:
        """Site directory.

        This directory, which is part of the theme's package, contains the
        files and subdirectories that define the built site.

        Files that have a ``.jinja`` extension are treated as Jinja templates
        and their contents are rendered.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def templates_dir(self) -> Path:
        """Templates directory.

        This directory, which is part of the theme's package, contains Jinja
        templates that are included by either Jinja-templates files, or
        other templates.
        """
        raise NotImplementedError

    def _write_metadata(self, output_dir: Path) -> None:
        metadata_json = self.metadata.json()
        metadata_path = output_dir.joinpath("metadata.json")
        metadata_path.write_text(metadata_json)



"""Content extraction form TeX source"""

from __future__ import annotations

import logging
import re
from dataclasses import dataclass
from typing import Dict, Iterator, List, Optional, Union

__all__ = [
    "get_macros",
    "get_def_macros",
    "get_newcommand_macros",
    "LaTeXCommandElement",
    "LaTeXCommand",
    "ParsedCommand",
    "ParsedElement",
]

# Regular expressions for "\def \name {content}"
# Expects the entire command to be on one line.
DEF_PATTERN = re.compile(
    r"\\def\s*"  # def command with optional whitespace
    r"(?P<name>\\[a-zA-Z]*?)\s*"  # macro name with optional whitespace
    r"{(?P<content>.*?)}"
)  # macro contents


def get_macros(tex_source: str) -> Dict[str, str]:
    r"""Get all macro definitions from TeX source, supporting multiple
    declaration patterns.

    Parameters
    ----------
    tex_source : `str`
        TeX source content.

    Returns
    -------
    macros : `dict`
        Keys are macro names (including leading ``\``) and values are the
        content (as `str`) of the macros.

    Notes
    -----
    This function uses the following function to scrape macros of different
    types:

    - `get_def_macros`
    - `get_newcommand_macros`

    This macro scraping has the following caveats:

    - Macro definition (including content) must all occur on one line.
    - Macros with arguments are not supported.
    """
    macros = {}
    macros.update(get_def_macros(tex_source))
    macros.update(get_newcommand_macros(tex_source))
    return macros


def get_def_macros(tex_source: str) -> Dict[str, str]:
    r"""Get all ``\def`` macro definition from TeX source.

    Parameters
    ----------
    tex_source : `str`
        TeX source content.

    Returns
    -------
    macros : `dict`
        Keys are macro names (including leading ``\``) and values are the
        content (as `str`) of the macros.

    Notes
    -----
    ``\def`` macros with arguments are not supported.
    """
    macros = {}
    for match in DEF_PATTERN.finditer(tex_source):
        macros[match.group("name")] = match.group("content")
    return macros


def get_newcommand_macros(tex_source: str) -> Dict[str, str]:
    r"""Get all ``\newcommand`` macro definition from TeX source.

    Parameters
    ----------
    tex_source : `str`
        TeX source content.

    Returns
    -------
    macros : `dict`
        Keys are macro names (including leading ``\``) and values are the
        content (as `str`) of the macros.

    Notes
    -----
    ``\newcommand`` macros with arguments are not supported.
    """
    macros = {}
    command = LaTeXCommand(
        "newcommand",
        LaTeXCommandElement(name="name", required=True, bracket="{"),
        LaTeXCommandElement(name="content", required=True, bracket="{"),
    )

    for macro in command.parse(tex_source):
        macros[macro["name"]] = macro["content"]

    return macros


@dataclass
class LaTeXCommandElement:

    bracket: Optional[str] = None
    """The element's bracket style, either ``[`` or ``{``."""

    required: bool = False
    """Whether the element is considered optional in the command invocation.
    """

    name: Optional[str] = None
    """Optional name of the element."""

    index: Optional[int] = None
    """Index of the command element in the command."""


@dataclass
class ParsedElement:

    name: Optional[str]
    """Optional name of the element (`None` if the element is not named)."""

    index: int
    """Index of the command element in the command."""

    content: str
    """Content of the command element (tex source)."""


class LaTeXCommand:
    r"""Definition of a LaTeX command's syntax that is used for parsing that
    command's content from a LaTeX source document.

    Parameters
    ----------
    name
        Name of the LaTeX command. For example, the name of the ``'\title'``
        command is ``'title'`` (without the backslash prefix).
    *elements
        Each position element is a dictionary describing that element of the
        command. These elements are provided in the order they appear in
        the command syntax.

        The keys of each dictionary are:

        - ``bracket``: The element's bracket style. Can be ``'['`` or ``'{'``.
        - ``required`` (optional field): `False` if the field is optional.
          `True` if required. Default is `True`.
        - ``name`` (optional field): Name of the field.
    """

    _brackets: Dict[str, str] = {"[": "]", "{": "}"}

    def __init__(self, name: str, *elements: LaTeXCommandElement) -> None:
        super().__init__()
        self._logger = logging.getLogger(__name__)
        self.name = name
        # Should add validation to elements
        self.elements: List[LaTeXCommandElement] = []
        for i, element in enumerate(elements):
            element.index = i
            self.elements.append(element)

    def parse(self, source: str) -> Iterator[ParsedCommand]:
        """Parse command content from the LaTeX source.

        Parameters
        ----------
        source
            The full source of the tex document.

        Yields
        ------
        parsed_command : `ParsedCommand`
            Yields parsed commands instances for each occurence of the command
            in the source.
        """
        command_regex = self._make_command_regex(self.name)
        for match in re.finditer(command_regex, source):
            self._logger.debug(match)
            start_index = match.start(0)
            yield self._parse_command(source, start_index)

    @staticmethod
    def _make_command_regex(name: str) -> str:
        r"""Given a command name, build a regular expression to detect that
        command in TeX source.

        The regular expression is designed to discern "\title{..}" from
        "\titlename{..}". It does this by ensuring that the command is
        followed by a whitespace character, argument brackets, or a comment
        character.

        Parameters
        ----------
        name
            Name of the command (with a backslash prefix).

        Returns
        -------
        regex
            Regular expression pattern for detecting the command.
        """
        return r"\\" + name + r"(?:[\s{[%])"

    def _parse_command(self, source: str, start_index: int) -> ParsedCommand:
        """Parse a single command.

        Parameters
        ----------
        source : `str`
            The full source of the tex document.
        start_index : `int`
            Character index in ``source`` where the command begins.

        Returns
        -------
        parsed_command
            The parsed command from the source at the given index.
        """
        parsed_elements = []

        # Index of the parser in the source
        running_index = start_index

        for element in self.elements:
            assert element.bracket is not None
            opening_bracket = element.bracket
            closing_bracket = self._brackets[opening_bracket]

            # Find the opening bracket.
            element_start = None
            element_end = None
            for i, c in enumerate(source[running_index:], start=running_index):
                if c == element.bracket:
                    element_start = i
                    break
                elif c == "\n":
                    # No starting bracket on the line.
                    if element.required is True:
                        # Try to parse a single single-word token after the
                        # command, like '\input file'
                        content = self._parse_whitespace_argument(
                            source[running_index:], self.name
                        )
                        assert element.index is not None
                        return ParsedCommand(
                            self.name,
                            [
                                ParsedElement(
                                    index=element.index,
                                    name=element.name,
                                    content=content.strip(),
                                )
                            ],
                            start_index,
                            source[start_index:i],
                        )
                    else:
                        # Give up on finding an optional element
                        break

            # Handle cases when the opening bracket is never found.
            if element_start is None and element.required is False:
                # Optional element not found. Continue to next element,
                # not advancing the running_index of the parser.
                continue
            elif element_start is None and element.required is True:
                message = (
                    f"Parsing command {self.name} at index {start_index:d}, "
                    f"did not detect element {element.index:d}"
                )
                raise RuntimeError(message)

            # Find the closing bracket, keeping track of the number of times
            # the same type of bracket was opened and closed.
            balance = 1
            assert element_start is not None
            for i, c in enumerate(
                source[element_start + 1 :], start=element_start + 1
            ):
                if c == opening_bracket:
                    balance += 1
                elif c == closing_bracket:
                    balance -= 1

                if balance == 0:
                    element_end = i
                    break

            if balance > 0:
                message = (
                    f"Parsing command {self.name} at index {start_index:d}, "
                    "did not find closing bracket for required "
                    f"command element {element.index:d}"
                )
                raise RuntimeError(message)

            # Package the parsed element's content.
            element_content = source[element_start + 1 : element_end]
            assert element.index is not None
            assert element_end is not None
            parsed_element = ParsedElement(
                index=element.index,
                name=element.name,
                content=element_content.strip(),
            )
            parsed_elements.append(parsed_element)

            running_index = element_end + 1

        command_source = source[start_index:running_index]
        parsed_command = ParsedCommand(
            self.name, parsed_elements, start_index, command_source
        )
        return parsed_command

    @staticmethod
    def _parse_whitespace_argument(source: str, name: str) -> str:
        r"""Attempt to parse a single token on the first line of this source.

        This method is used for parsing whitespace-delimited arguments, like
        ``\input file``. The source should ideally contain `` file`` along
        with a newline character.

        >>> source = 'Line 1\n' r'\input test.tex' '\nLine 2'
        >>> LatexCommand._parse_whitespace_argument(source, 'input')
        'test.tex'

        Bracket delimited arguments (``\input{test.tex}``) are handled in
        the normal logic of `_parse_command`.
        """
        # First match the command name itself so that we find the argument
        # *after* the command
        command_pattern = r"\\(" + name + r")(?:[\s{[%])"
        command_match = re.search(command_pattern, source)
        if command_match is not None:
            # Trim `source` so we only look after the command
            source = source[command_match.end(1) :]

        # Find the whitespace-delimited argument itself.
        pattern = r"(?P<content>\S+)(?:[ %\t\n]+)"
        match = re.search(pattern, source)
        if match is None:
            message = (
                "When parsing {}, did not find whitespace-delimited command "
                "argument"
            )
            raise RuntimeError(message.format(name))
        content = match.group("content")
        content.strip()
        return content


class ParsedCommand:
    r"""Contents of a parsed LaTeX command.

    Parameters
    ----------
    name
        Name of the LaTeX command. For example, the name of the ``'\title'``
        command is ``'title'`` (without the backslash prefix).
    parsed_elements
        Parsed command elements.
    start_index
        Character index in the ``full_source`` where the command begins.
    command_source
        The full source of the parsed LaTeX command. This can be used to
        replace the full command in the source document.
    """

    def __init__(
        self,
        name: str,
        parsed_elements: List[ParsedElement],
        start_index: int,
        command_source: str,
    ) -> None:
        self.name: str = name
        self.start_index: int = start_index
        self._parsed_elements: List[ParsedElement] = parsed_elements
        self.command_source: str = command_source

    def __getitem__(self, key: Union[int, str]) -> str:
        element = self._get_element(key)
        return element.content

    def __contains__(self, key: Union[int, str]) -> bool:
        try:
            self._get_element(key)
        except KeyError:
            return False

        return True

    def _get_element(self, key: Union[int, str]) -> ParsedElement:
        element = None

        if isinstance(key, int):
            # Get by element index
            for _element in self._parsed_elements:
                if _element.index == key:
                    element = _element
                    break
        else:
            # Get by element name
            for _element in self._parsed_elements:
                if _element.name == key:
                    element = _element
                    break

        if element is None:
            message = "Key {} not found".format(key)
            raise KeyError(message)
        return element



from __future__ import annotations

from typing import Any, Dict

from lander.ext.parser import DocumentMetadata, Parser
from lander.ext.parser.texutils.extract import (
    LaTeXCommand,
    LaTeXCommandElement,
)

__all__ = ["ArticleParser"]


class ArticleParser(Parser):
    """A parser for LaTeX article class documents."""

    def extract_metadata(self) -> DocumentMetadata:
        """Extract meta from an article document (parser hook).

        Returns
        -------
        metadata
            The metadata parsed from the document source.
        """
        metadata: Dict[str, Any] = {}
        try:
            metadata["date_modified"] = self._parse_date(self.tex_source)
        except RuntimeError:
            pass
        metadata["title"] = self._parse_title(self.tex_source)

        # Apply canonical URL setting, if available
        if self.settings.canonical_url:
            metadata["canonical_url"] = self.settings.canonical_url

        # Override metadata from YAML or CLI as available
        metadata.update(self.settings.metadata)

        return DocumentMetadata(**metadata)

    def _parse_title(self, tex_source: str) -> str:
        command = LaTeXCommand(
            "title",
            LaTeXCommandElement(
                name="short_title", required=False, bracket="["
            ),
            LaTeXCommandElement(name="long_title", required=True, bracket="{"),
        )
        titles = [t for t in command.parse(tex_source)]
        if len(titles) == 0:
            raise RuntimeError("Could not parse a title command.")
        return titles[-1]["long_title"]

    def _parse_date(self, tex_source: str) -> str:
        command = LaTeXCommand(
            "date",
            LaTeXCommandElement(name="date", required=True, bracket="{"),
        )
        dates = [t for t in command.parse(tex_source)]
        if len(dates) == 0:
            raise RuntimeError("Could not parse a date command.")
        return dates[-1]["date"]



"""Metadata retrieval from a local Git repository."""

from __future__ import annotations

import datetime
from dataclasses import dataclass
from pathlib import Path, PurePath
from typing import List, Optional, Sequence

import git

__all__ = ["GitFile", "GitRepository"]


@dataclass
class GitFile:
    """A file in Git."""

    path: Path

    name: PurePath

    date_modified: datetime.datetime

    @property
    def extension(self) -> str:
        """The file's extension (without the ``.`` suffix)."""
        return self.name.suffix[1:]  # remove "."


@dataclass
class GitRepository:
    """A Git repository."""

    path: Path
    """Path for the root of the repository."""

    date_modified: datetime.datetime
    """Date when the head commit was created."""

    repo: git.Repo
    """The GitPython repository representation."""

    files: List[GitFile]
    """"Files in the Git repository."""

    @classmethod
    def create(cls, path: Path) -> GitRepository:
        """Create a GitRepository from a Git repository located on the
        file system.
        """
        repo = git.Repo(path=str(path), search_parent_directories=True)
        repo_path = Path(repo.working_tree_dir)
        head_commit = repo.head.commit

        files = GitRepository._gather_files(repo)

        return cls(
            path=repo_path,
            date_modified=head_commit.committed_datetime,
            repo=repo,
            files=files,
        )

    @staticmethod
    def _gather_files(repo: git.Repo) -> List[GitFile]:
        """Gather metadata about all files in the Git tree."""
        files: List[GitFile] = []

        head_commit = repo.head.commit

        for item in head_commit.tree.traverse():
            if item.type == "blob":
                try:
                    git_file = GitRepository._import_file(
                        filepath=Path(item.abspath), repo=repo
                    )
                    files.append(git_file)
                except IOError:
                    continue

        return files

    @staticmethod
    def _import_file(*, filepath: Path, repo: git.Repo) -> GitFile:
        """Gather Git metadata about a file.

        Parameters
        ----------
        filepath : `Path`
            Absolute or repository-relative path for a file.
        repo : `git.Repo`
            A `git.Repo` instance.

        Returns
        -------
        git_file : `GitFile`
            Representation of the Git file.

        Raises
        ------
        IOError
            Raised if the ``filepath`` does not exist in the Git repository.
        """
        head_commit = repo.head.commit

        # Filepath relative to the repo path
        repo_path = Path(repo.working_tree_dir)
        repo_rel_path = filepath.resolve().relative_to(repo_path.resolve())

        # Most recent commit datetime of the given file.
        # Don't use head_commit.iter_parents because then it skips the
        # commit of a file that's added but never modified.
        for commit in head_commit.iter_items(
            repo, head_commit, [repo_rel_path], skip=0,
        ):
            return GitFile(
                path=filepath.resolve(),
                name=PurePath(repo_rel_path),
                date_modified=commit.committed_datetime,
            )

        # Only get here if git could not find the file path in the history
        raise IOError("File {} not found".format(filepath))

    def compute_date_modified(
        self, extensions: Optional[Sequence[str]] = None
    ) -> datetime.datetime:
        """Get the most recent modification date, optional considering only
        files with one of an accepted sequence of extensions.

        Parameters
        ----------
        extensions : sequence of `str`
            Extension names, such as ``"pdf"``, ``"tex"``.
        """
        dates_modified = []
        for git_file in self.files:
            if extensions and git_file.extension not in extensions:
                continue
            dates_modified.append(git_file.date_modified)

        return max(dates_modified)
